#!/usr/bin/env python3
"""Fast (?) and accurate elliptic integral of the second kind.

Reference:
    - https://github.com/stdlib-js/math-base-special-ellipk/blob/main/src/main.c
    - Fukushima, Toshio. 2009. "Fast computation of complete elliptic integrals and Jacobian elliptic functions."
    - Fukushima, Toshio. 2015. "Precise and fast computation of complete elliptic integrals by piecewise minimax rational function approximation."
"""
import torch
from scipy.special import ellipe as s_ellipe
from torch import Tensor

from pymytools.special._ellipk import ellipk


def poly_p1(x: Tensor):
    return 1.5509733517804722 + (
        x
        * (
            -0.4003010201031985
            + (
                x
                * (
                    -0.07849861944294194
                    + (
                        x
                        * (
                            -0.034318853117591995
                            + (
                                x
                                * (
                                    -0.0197180433173655
                                    + (
                                        x
                                        * (
                                            -0.01305950773199331
                                            + (
                                                x
                                                * (
                                                    -0.009442372874146548
                                                    + (
                                                        x
                                                        * (
                                                            -0.007246728512402157
                                                            + (
                                                                x
                                                                * (
                                                                    -0.00580742401295609
                                                                    + (
                                                                        x
                                                                        * -0.004809187786009338
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p2(x: Tensor):
    return 1.5101218320928198 + (
        x
        * (
            -0.41711633390586755
            + (
                x
                * (
                    -0.09012382040477457
                    + (
                        x
                        * (
                            -0.04372994401908431
                            + (
                                x
                                * (
                                    -0.027965493064761784
                                    + (
                                        x
                                        * (
                                            -0.020644781177568104
                                            + (
                                                x
                                                * (
                                                    -0.016650786739707237
                                                    + (
                                                        x
                                                        * (
                                                            -0.01426196082884252
                                                            + (
                                                                x
                                                                * (
                                                                    -0.012759847429264804
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -0.011799303775587354
                                                                            + (
                                                                                x
                                                                                * -0.011197445703074968
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p3(x: Tensor):
    return 1.4674622093394272 + (
        x
        * (
            -0.43657629094633776
            + (
                x
                * (
                    -0.10515555766694255
                    + (
                        x
                        * (
                            -0.05737184359324173
                            + (
                                x
                                * (
                                    -0.04139162772734022
                                    + (
                                        x
                                        * (
                                            -0.03452772850528084
                                            + (
                                                x
                                                * (
                                                    -0.031495443512532785
                                                    + (
                                                        x
                                                        * (
                                                            -0.030527000890325277
                                                            + (
                                                                x
                                                                * (
                                                                    -0.0309169840192389
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -0.03237139531475812
                                                                            + (
                                                                                x
                                                                                * -0.03478996038640416
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p4(x: Tensor):
    return 1.4226911334908792 + (
        x
        * (
            -0.4595135196210487
            + (
                x
                * (
                    -0.12525053982206188
                    + (
                        x
                        * (
                            -0.07813854509440948
                            + (
                                x
                                * (
                                    -0.06471427847205
                                    + (
                                        x
                                        * (
                                            -0.06208433913173031
                                            + (
                                                x
                                                * (
                                                    -0.06519703281557247
                                                    + (
                                                        x
                                                        * (
                                                            -0.07279389536257878
                                                            + (
                                                                x
                                                                * (
                                                                    -0.084959075171781
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -0.102539850131046
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -0.12705358515769605
                                                                                    + (
                                                                                        x
                                                                                        * -0.1607911206912746
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p5(x: Tensor):
    return 1.3754019718711163 + (
        x
        * (
            -0.4872021832731848
            + (
                x
                * (
                    -0.15331170134854022
                    + (
                        x
                        * (
                            -0.11184944491702783
                            + (
                                x
                                * (
                                    -0.10884095252313576
                                    + (
                                        x
                                        * (
                                            -0.12295422312026907
                                            + (
                                                x
                                                * (
                                                    -0.15221716396203505
                                                    + (
                                                        x
                                                        * (
                                                            -0.20049532364269734
                                                            + (
                                                                x
                                                                * (
                                                                    -0.27617433306775174
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -0.39351311430437586
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -0.5757544060278792
                                                                                    + (
                                                                                        x
                                                                                        * (
                                                                                            -0.8605232357272398
                                                                                            + (
                                                                                                x
                                                                                                * -1.3088332057585401
                                                                                            )
                                                                                        )
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p6(x: Tensor):
    return 1.3250244979582302 + (
        x
        * (
            -0.5217276475575667
            + (
                x
                * (
                    -0.19490643048212622
                    + (
                        x
                        * (
                            -0.17162372682201127
                            + (
                                x
                                * (
                                    -0.20275465292641914
                                    + (
                                        x
                                        * (
                                            -0.27879895311853475
                                            + (
                                                x
                                                * (
                                                    -0.42069845728100574
                                                    + (
                                                        x
                                                        * (
                                                            -0.675948400853106
                                                            + (
                                                                x
                                                                * (
                                                                    -1.1363431218392293
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -1.9767211439543984
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -3.5316967730957227
                                                                                    + (
                                                                                        x
                                                                                        * (
                                                                                            -6.446753640156048
                                                                                            + (
                                                                                                x
                                                                                                * -11.97703130208884
                                                                                            )
                                                                                        )
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p7(x: Tensor):
    return 1.2707074796501499 + (
        x
        * (
            -0.5668391682878666
            + (
                x
                * (
                    -0.2621607934324926
                    + (
                        x
                        * (
                            -0.2922441735330774
                            + (
                                x
                                * (
                                    -0.4403978408504232
                                    + (
                                        x
                                        * (
                                            -0.7749476413813975
                                            + (
                                                x
                                                * (
                                                    -1.498870837987561
                                                    + (
                                                        x
                                                        * (
                                                            -3.089708310445187
                                                            + (
                                                                x
                                                                * (
                                                                    -6.6675959033810015
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -14.89436036517319
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -34.18120574251449
                                                                                    + (
                                                                                        x
                                                                                        * (
                                                                                            -80.15895841905397
                                                                                            + (
                                                                                                x
                                                                                                * (
                                                                                                    -191.34894807629848
                                                                                                    + (
                                                                                                        x
                                                                                                        * (
                                                                                                            -463.5938853480342
                                                                                                            + (
                                                                                                                x
                                                                                                                * -1137.38082216936
                                                                                                            )
                                                                                                        )
                                                                                                    )
                                                                                                )
                                                                                            )
                                                                                        )
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p8(x: Tensor):
    return 1.2110560275684594 + (
        x
        * (
            -0.6303064132874558
            + (
                x
                * (
                    -0.38716640952066916
                    + (
                        x
                        * (
                            -0.5922782353119346
                            + (
                                x
                                * (
                                    -1.23755558451305
                                    + (
                                        x
                                        * (
                                            -3.0320566617452474
                                            + (
                                                x
                                                * (
                                                    -8.18168822157359
                                                    + (
                                                        x
                                                        * (
                                                            -23.55507217389693
                                                            + (
                                                                x
                                                                * (
                                                                    -71.04099935893065
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -221.879685319235
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -712.1364793277636
                                                                                    + (
                                                                                        x
                                                                                        * (
                                                                                            -2336.1253314403966
                                                                                            + (
                                                                                                x
                                                                                                * (
                                                                                                    -7801.945954775964
                                                                                                    + (
                                                                                                        x
                                                                                                        * (
                                                                                                            -26448.19586059192
                                                                                                            + (
                                                                                                                x
                                                                                                                * (
                                                                                                                    -90799.48341621365
                                                                                                                    + (
                                                                                                                        x
                                                                                                                        * (
                                                                                                                            -315126.04064491636
                                                                                                                            + (
                                                                                                                                x
                                                                                                                                * -1104011.3443115912
                                                                                                                            )
                                                                                                                        )
                                                                                                                    )
                                                                                                                )
                                                                                                            )
                                                                                                        )
                                                                                                    )
                                                                                                )
                                                                                            )
                                                                                        )
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p9(x: Tensor):
    return 1.1613071521962828 + (
        x
        * (
            -0.7011002845552895
            + (
                x
                * (
                    -0.5805514744654373
                    + (
                        x
                        * (
                            -1.2436930610777865
                            + (
                                x
                                * (
                                    -3.679383613496635
                                    + (
                                        x
                                        * (
                                            -12.815909243378957
                                            + (
                                                x
                                                * (
                                                    -49.25672530759985
                                                    + (
                                                        x
                                                        * (
                                                            -202.18187354340904
                                                            + (
                                                                x
                                                                * (
                                                                    -869.8602699308701
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -3877.0058473132895
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -17761.7071017094
                                                                                    + (
                                                                                        x
                                                                                        * (
                                                                                            -83182.69029154233
                                                                                            + (
                                                                                                x
                                                                                                * (
                                                                                                    -396650.4505013548
                                                                                                    + (
                                                                                                        x
                                                                                                        * -1920033.4136826345
                                                                                                    )
                                                                                                )
                                                                                            )
                                                                                        )
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p10(x: Tensor):
    return 1.1246173251197522 + (
        x
        * (
            -0.7708450563609095
            + (
                x
                * (
                    -0.8447940536449113
                    + (
                        x
                        * (
                            -2.4900973094503946
                            + (
                                x
                                * (
                                    -10.239717411543843
                                    + (
                                        x
                                        * (
                                            -49.7490054655148
                                            + (
                                                x
                                                * (
                                                    -267.09866751957054
                                                    + (
                                                        x
                                                        * (
                                                            -1532.66588382523
                                                            + (
                                                                x
                                                                * (
                                                                    -9222.313478526092
                                                                    + (
                                                                        x
                                                                        * (
                                                                            -57502.51612140314
                                                                            + (
                                                                                x
                                                                                * (
                                                                                    -368596.11674161063
                                                                                    + (
                                                                                        x
                                                                                        * (
                                                                                            -2415611.0887010912
                                                                                            + (
                                                                                                x
                                                                                                * (
                                                                                                    -16120097.815816568
                                                                                                    + (
                                                                                                        x
                                                                                                        * (
                                                                                                            -109209938.52030899
                                                                                                            + (
                                                                                                                x
                                                                                                                * (
                                                                                                                    -749380758.1942496
                                                                                                                    + (
                                                                                                                        x
                                                                                                                        * (
                                                                                                                            -5198725846.725541
                                                                                                                            + (
                                                                                                                                x
                                                                                                                                * -36409256888.1214
                                                                                                                            )
                                                                                                                        )
                                                                                                                    )
                                                                                                                )
                                                                                                            )
                                                                                                        )
                                                                                                    )
                                                                                                )
                                                                                            )
                                                                                        )
                                                                                    )
                                                                                )
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p11(x: Tensor):
    return 1.5910034537907922 + (
        x
        * (
            0.41600074399178694
            + (
                x
                * (
                    0.24579151426410342
                    + (
                        x
                        * (
                            0.17948148291490615
                            + (
                                x
                                * (
                                    0.14455605708755515
                                    + (
                                        x
                                        * (
                                            0.12320099331242772
                                            + (
                                                x
                                                * (
                                                    0.10893881157429353
                                                    + (
                                                        x
                                                        * (
                                                            0.09885340987159291
                                                            + (
                                                                x
                                                                * (
                                                                    0.09143962920174975
                                                                    + (
                                                                        x
                                                                        * (
                                                                            0.0858425915954139
                                                                            + (
                                                                                x
                                                                                * 0.08154111871830322
                                                                            )
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def poly_p12(x: Tensor):
    return 1.5509733517804722 + (
        x
        * (
            -0.4003010201031985
            + (
                x
                * (
                    -0.07849861944294194
                    + (
                        x
                        * (
                            -0.034318853117591995
                            + (
                                x
                                * (
                                    -0.0197180433173655
                                    + (
                                        x
                                        * (
                                            -0.01305950773199331
                                            + (
                                                x
                                                * (
                                                    -0.009442372874146548
                                                    + (
                                                        x
                                                        * (
                                                            -0.007246728512402157
                                                            + (
                                                                x
                                                                * (
                                                                    -0.00580742401295609
                                                                    + (
                                                                        x
                                                                        * -0.004809187786009338
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def ellipe(m: Tensor) -> Tensor:
    """Computation of complete elliptic integral of the second kind. Modified version of `stdlib.js`'s `ellipk` to utilize `torch.vmap` functionality.

    Note:
        - Somehow, performance of this method is floor than that of SciPy. (even together with `vmap`)
        - Limitation of `vmap`
            - Conditional flow is not allowed
            - Tensor with indices is not working properly
            - `vmap` over `scipy` function is not working
        - Can I improve this method?
    """

    x = m.clone()

    x = torch.where(m < 0.0, m / (m - 1.0), x)

    x = torch.where(x == 0.0, torch.pi / 2.0, x)
    x = torch.where(x == 1.0, 1.0, x)
    x = torch.where(x > 1.0, torch.nan, x)

    x = torch.where((x > 0.0) & (x < 0.1), poly_p1(x - 0.05), x)
    x = torch.where((x >= 0.1) & (x < 0.2), poly_p2(x - 0.15), x)
    x = torch.where((x >= 0.2) & (x < 0.3), poly_p3(x - 0.25), x)
    x = torch.where((x >= 0.3) & (x < 0.4), poly_p4(x - 0.35), x)
    x = torch.where((x >= 0.4) & (x < 0.5), poly_p5(x - 0.45), x)
    x = torch.where((x >= 0.5) & (x < 0.6), poly_p6(x - 0.55), x)
    x = torch.where((x >= 0.6) & (x < 0.7), poly_p7(x - 0.65), x)
    x = torch.where((x >= 0.7) & (x < 0.8), poly_p8(x - 0.75), x)
    x = torch.where((x >= 0.8) & (x < 0.85), poly_p9(x - 0.825), x)
    x = torch.where((x >= 0.85) & (x < 0.9), poly_p10(x - 0.875), x)
    x = torch.where(
        (x >= 0.9) & (x < 1.0),
        (torch.pi / 2 + ellipk(x) * (poly_p11(0.95 - x) - poly_p12(0.95 - x)))
        / poly_p11(0.95 - x),
        x,
    )
    x = torch.where(m < 0.0, x * torch.sqrt(1.0 - m), x)

    return x


def t_ellipe(x: Tensor):
    """Wrapper for scipy.special.ellipe."""
    return s_ellipe(x.cpu()).to(x.device, x.dtype)
